@@ -4,8 +4,6 @@
 #include "Imlib_private.h"
 #include <setjmp.h>
 
-#define G_MAXINT ((int) 0x7fffffff)
-
 /*      Split the ID - damages input    */
 
 static char        *
@@ -43,17 +41,13 @@ _GetExtension(char *file)
 
 /*
  *     Make sure we don't wrap on our memory allocations
- *     we check G_MAXINT/4 because rend.c malloc's w * h * bpp
- *     + 3 is safety margin
  */
 
 void * _imlib_malloc_image(unsigned int w, unsigned int h)
 {
-       if (w <= 0 || w > 32767 ||
-           h <= 0 || h > 32767 ||
-           h >= (G_MAXINT/4 - 1) / w)
+       if( w > 32767 || h > 32767)
                return NULL;
-       return malloc(w * h * 3 + 3);
+       return malloc(w * h * 3);
 }
 
 #ifdef HAVE_LIBJPEG
@@ -366,9 +360,7 @@ _LoadTIFF(ImlibData * id, FILE *f, char *file, int *w, int *h, int *t)
   npix = ww * hh;
   *w = (int)ww;
   *h = (int)hh;
-  if (ww <= 0 || ww > 32767 ||
-      hh <= 0 || hh > 32767 ||
-      hh >= (G_MAXINT/sizeof(uint32)) / ww)
+  if(ww > 32767 || hh > 32767)
     {
        TIFFClose(tif);
        return NULL;
@@ -471,7 +463,7 @@ _LoadGIF(ImlibData * id, FILE *f, int *w, int *h, int *t)
 	    }
 	  *w = gif->Image.Width;
 	  *h = gif->Image.Height;
-	  if (*h <= 0 || *h > 32767 || *w <= 0 || *w > 32767)
+	  if (*h > 32767 || *w > 32767)
 	    {
 	       return NULL;
 	    }
@@ -639,12 +631,12 @@ _LoadBMP(ImlibData * id, FILE *file, int *w, int *h, int *t)
   fread(dbuf, 4, 2, file);
   *w = (int)dbuf[0];
   *h = (int)dbuf[1];
-  if (*w > 32767)
+  if ((*w > 32767) || (*w < 0))
     {
       fprintf(stderr, "IMLIB ERROR: Image width > 32767 pixels for file\n");
       return NULL;
     }
-  if (*h > 32767)
+  if ((*h > 32767) || (*h < 0))
     {
       fprintf(stderr, "IMLIB ERROR: Image height > 32767 pixels for file\n");
       return NULL;
@@ -667,8 +659,11 @@ _LoadBMP(ImlibData * id, FILE *file, int *w, int *h, int *t)
     }
   fread(dbuf, 4, 2, file);
   ncolors = (int)dbuf[0];
-  if (ncolors == 0)
+  if ((ncolors > (1 << bpp)) || (ncolors < 0))
     ncolors = 1 << bpp;
+  if ((ncolors < 0) || (ncolors > (1 << bpp)))
+    ncolors = 1 << bpp;
+
   /* some more sanity checks */
   if (((comp == BI_RLE4) && (bpp != 4)) || ((comp == BI_RLE8) && (bpp != 8)) || ((comp == BI_BITFIELDS) && (bpp != 16 && bpp != 32)))
     {
@@ -794,9 +789,13 @@ _LoadBMP(ImlibData * id, FILE *file, int *w, int *h, int *t)
 		  for (bit = 0; bit < 8; bit++)
 		    {
 		      index = ((byte & (0x80 >> bit)) ? 1 : 0);
-		      ptr[poffset] = cmap[index].r;
-		      ptr[poffset + 1] = cmap[index].g;
-		      ptr[poffset + 2] = cmap[index].b;
+		      /* possibly corrupted file? */
+		      if (index < ncolors && poffset < *w * *h * 3)
+			{
+			  ptr[poffset] = cmap[index].r;
+			  ptr[poffset + 1] = cmap[index].g;
+			  ptr[poffset + 2] = cmap[index].b;
+			}
 		      column++;
 		    }
 		}
@@ -818,9 +817,13 @@ _LoadBMP(ImlibData * id, FILE *file, int *w, int *h, int *t)
 			  index = ((byte & (0xF0 >> nibble * 4)) >> (!nibble * 4));
 			  if (index >= 16)
 			    index = 15;
-			  ptr[poffset] = cmap[index].r;
-			  ptr[poffset + 1] = cmap[index].g;
-			  ptr[poffset + 2] = cmap[index].b;
+			  /* possibly corrupted file? */
+			  if (index < ncolors && poffset < *w * *h * 3)
+			    {
+			      ptr[poffset] = cmap[index].r;
+			      ptr[poffset + 1] = cmap[index].g;
+			      ptr[poffset + 2] = cmap[index].b;
+			    }
 			  column++;
 			}
 		    }
@@ -859,9 +862,13 @@ _LoadBMP(ImlibData * id, FILE *file, int *w, int *h, int *t)
 				{
 				  linepos++;
 				  byte = getc(file);
-				  ptr[poffset] = cmap[byte].r;
-				  ptr[poffset + 1] = cmap[byte].g;
-				  ptr[poffset + 2] = cmap[byte].b;
+				  /* possibly corrupted file? */
+				  if (byte < ncolors && poffset < *w * *h * 3)
+				    {
+				      ptr[poffset] = cmap[byte].r;
+				      ptr[poffset + 1] = cmap[byte].g;
+				      ptr[poffset + 2] = cmap[byte].b;
+				    }
 				  column++;
 				}
 			      if (absolute & 0x01)
@@ -872,9 +879,13 @@ _LoadBMP(ImlibData * id, FILE *file, int *w, int *h, int *t)
 			{
 			  for (i = 0; i < first; i++)
 			    {
-			      ptr[poffset] = cmap[byte].r;
-			      ptr[poffset + 1] = cmap[byte].g;
-			      ptr[poffset + 2] = cmap[byte].b;
+			      /* possibly corrupted file? */
+			      if (byte < ncolors && poffset < *w * *h * 3)
+				{
+				  ptr[poffset] = cmap[byte].r;
+				  ptr[poffset + 1] = cmap[byte].g;
+				  ptr[poffset + 2] = cmap[byte].b;
+				}
 			      column++;
 			      linepos++;
 			    }
@@ -882,9 +893,13 @@ _LoadBMP(ImlibData * id, FILE *file, int *w, int *h, int *t)
 		    }
 		  else
 		    {
-		      ptr[poffset] = cmap[byte].r;
-		      ptr[poffset + 1] = cmap[byte].g;
-		      ptr[poffset + 2] = cmap[byte].b;
+		      /* possibly corrupted file? */
+		      if (byte < ncolors && poffset < *w * *h * 3)
+			{
+			  ptr[poffset] = cmap[byte].r;
+			  ptr[poffset + 1] = cmap[byte].g;
+			  ptr[poffset + 2] = cmap[byte].b;
+			}
 		      column++;
 		    }
 		}
@@ -892,9 +907,13 @@ _LoadBMP(ImlibData * id, FILE *file, int *w, int *h, int *t)
 	  else if (bpp == 24)
 	    {
 	      linepos += fread(bbuf, 1, 3, file);
-	      ptr[poffset] = (unsigned char)bbuf[2];
-	      ptr[poffset + 1] = (unsigned char)bbuf[1];
-	      ptr[poffset + 2] = (unsigned char)bbuf[0];
+	      /* possibly corrupted file? */
+	      if (poffset < *w * *h * 3)
+		{
+		  ptr[poffset] = (unsigned char)bbuf[2];
+		  ptr[poffset + 1] = (unsigned char)bbuf[1];
+		  ptr[poffset + 2] = (unsigned char)bbuf[0];
+		}
 	      column++;
 	    }
 	  else if (bpp == 16)
@@ -902,12 +921,16 @@ _LoadBMP(ImlibData * id, FILE *file, int *w, int *h, int *t)
 	      unsigned char       temp;
 
 	      linepos += fread(&word, 2, 1, file);
-	      temp = (word & rmask) >> rshift;
-	      ptr[poffset] = temp;
-	      temp = (word & gmask) >> gshift;
-	      ptr[poffset + 1] = temp;
-	      temp = (word & bmask) >> gshift;
-	      ptr[poffset + 2] = temp;
+	      /* possibly corrupted file? */
+	      if (poffset < *w * *h * 3)
+		{
+		  temp = (word & rmask) >> rshift;
+		  ptr[poffset] = temp;
+		  temp = (word & gmask) >> gshift;
+		  ptr[poffset + 1] = temp;
+		  temp = (word & bmask) >> gshift;
+		  ptr[poffset + 2] = temp;
+		}
 	      column++;
 	    }
 	  else
@@ -915,12 +938,16 @@ _LoadBMP(ImlibData * id, FILE *file, int *w, int *h, int *t)
 	      unsigned char       temp;
 
 	      linepos += fread(&dword, 4, 1, file);
-	      temp = (dword & rmask) >> rshift;
-	      ptr[poffset] = temp;
-	      temp = (dword & gmask) >> gshift;
-	      ptr[poffset + 1] = temp;
-	      temp = (dword & bmask) >> bshift;
-	      ptr[poffset + 2] = temp;
+	      /* possibly corrupted file? */
+	      if (poffset < *w * *h * 3)
+		{
+		  temp = (dword & rmask) >> rshift;
+		  ptr[poffset] = temp;
+		  temp = (dword & gmask) >> gshift;
+		  ptr[poffset + 1] = temp;
+		  temp = (dword & bmask) >> bshift;
+		  ptr[poffset + 2] = temp;
+		}
 	      column++;
 	    }
 	}
@@ -973,12 +1000,7 @@ _LoadXPM(ImlibData * id, FILE *file, int *w, int *h, int *t)
   comment = 0;
   quote = 0;
   context = 0;
-  memset(lookup, 0, sizeof(lookup));
-
   line = malloc(lsz);
-  if (!line)
-    return NULL;
-
   while (!done)
     {
       pc = c;
@@ -1007,25 +1029,25 @@ _LoadXPM(ImlibData * id, FILE *file, int *w, int *h, int *t)
 		{
 		  /* Header */
 		  sscanf(line, "%i %i %i %i", w, h, &ncolors, &cpp);
-                  if (ncolors <= 0 || ncolors > 32766)
+                  if (ncolors > 32766)
 		    {
 		      fprintf(stderr, "IMLIB ERROR: XPM files wth colors > 32766 not supported\n");
 		      free(line);
 		      return NULL;
 		    }
-		  if (cpp <= 0 || cpp > 5)
+		  if (cpp > 5)
 		    {
 		      fprintf(stderr, "IMLIB ERROR: XPM files with characters per pixel > 5 not supported\n");
 		      free(line);
 		      return NULL;
 		    }
-		  if (*w <= 0 || *w > 32767)
+		  if (*w > 32767)
 		    {
 		      fprintf(stderr, "IMLIB ERROR: Image width > 32767 pixels for file\n");
 		      free(line);
 		      return NULL;
 		    }
-		  if (*h <= 0 || *h > 32767)
+		  if (*h > 32767)
 		    {
 		      fprintf(stderr, "IMLIB ERROR: Image height > 32767 pixels for file\n");
 		      free(line);
@@ -1058,13 +1080,11 @@ _LoadXPM(ImlibData * id, FILE *file, int *w, int *h, int *t)
 		    {
 		      int                 slen;
 		      int                 hascolor, iscolor;
-		      int                 space;
 
 		      iscolor = 0;
 		      hascolor = 0;
 		      tok[0] = 0;
 		      col[0] = 0;
-		      space = sizeof(col) - 1;
 		      s[0] = 0;
 		      len = strlen(line);
 		      strncpy(cmap[j].str, line, cpp);
@@ -1087,14 +1107,10 @@ _LoadXPM(ImlibData * id, FILE *file, int *w, int *h, int *t)
 				{
 				  if (k >= len)
 				    {
-				      if (col[0] && space > 0) {
+				      if (col[0])
 					strcat(col, " ");
-					space--;
-				      }
-				      if (slen <= space) {
+                                      if (strlen(col) + strlen(s) < sizeof(col))
 					strcat(col, s);
-					space -= slen;
-				      }
 				    }
 				  if (col[0])
 				    {
@@ -1124,22 +1140,14 @@ _LoadXPM(ImlibData * id, FILE *file, int *w, int *h, int *t)
 					    }
 					}
 				    }
-				  if (slen <= sizeof(tok))
-				    strcpy(tok, s);
+				  strcpy(tok, s);
 				  col[0] = 0;
-				  space = sizeof(col) - 1;
 				}
 			      else
 				{
-				  if (col[0] && space > 0)
-				  {
+				  if (col[0])
 				    strcat(col, " ");
-				    space--;
-				  }
-				  if (slen <= space) {
-				    strcat(col, s);
-				    space -= slen;
-				  }
+				  strcat(col, s);
 				}
 			    }
 			}
@@ -1368,12 +1376,12 @@ _LoadPPM(ImlibData * id, FILE * f, int *w, int *h)
 	  sscanf(s, "%i %i", w, h);
 	  a = *w;
 	  b = *h;
-	  if (a <= 0 || a > 32767)
+	  if (a > 32767)
 	    {
 	      fprintf(stderr, "IMLIB ERROR: Image width > 32767 pixels for file\n");
 	      return NULL;
 	    }
-	  if (b <= 0 || b > 32767)
+	  if (b > 32767)
 	    {
 	      fprintf(stderr, "IMLIB ERROR: Image height > 32767 pixels for file\n");
 	      return NULL;
@@ -1816,7 +1824,6 @@ Imlib_load_image(ImlibData * id, char *file)
     {
       fprintf(stderr, "IMLIB ERROR: Cannot load image: %s\n"
       		      "All fallbacks failed.\n", fil);
-      fprintf(stderr, "See /usr/share/doc/imlib2/README.fallback.\n");
       return NULL;
     }
     
