@@ -932,8 +932,11 @@ Imlib_init_with_params(Display * disp, ImlibInitParams * p)
     {
       if (p->flags & PARAMS_VISUALID)
 	vis = p->visualid;
-      if (p->flags & PARAMS_PALETTEFILE)
-	strcpy(palfile, p->palettefile);
+      if (p->flags & PARAMS_PALETTEFILE) {
+        if (palfile)
+            free(palfile);
+        palfile = strdup(p->palettefile);
+      }
       if (p->flags & PARAMS_SHAREDMEM)
 	{
 	  if (!p->sharedmem)
